#include <string>
#include <vector>
#include <cstdio>

/*
 * mkinterface
 * Reads a cproto-generated proto.h and modifies it to accomodate intermodule
 * stuff. In particular:
 *  - Any functions mentioned in <interface file> are made into a struct, and
 *    (if DLL_IMPORT is defined) references are defined to refer to the struct.
 *  - (For platforms requiring segmentation, when supported): Any functions
 *    mentioned in segment lists are marked accordingly
 */

FILE *out;

//{{{
class FunctionDefinition
{
public:
//{{{
	FunctionDefinition(const char *prototype)
	{
		// Parser definition:
		// A function's name is the token immediately preceding the first
		// open-parenthese. Everything prior is the type, and everything
		// after is the parameter list.
		
		const char *pos = prototype;
		const char *pos2;
		
		type = name = paramlist = std::string("");
		proto = std::string(prototype);
		is_function = true;
		
		if(prototype[0] == '/') // Hey, this is a comment line, not a proto
		{
			is_function = false;
			return;
		}
		
		while(*pos != '\0' && *pos != '(')
			pos++;
		if(*pos != '(') // This must not've been a prototype
		{
			is_function = false;
			return;
		}
		
		paramlist = std::string(pos);
		
		pos--;
		
		// Back up over whitespace
		while(isspace(*pos) && pos>prototype)
			pos--;
		// Now back up to the start of the name
		while( (isalpha(*pos)||*pos=='_') && pos>prototype )
			pos--;
		pos++;
		
		// Anything *before* this is the type
		for(pos2=prototype; pos2<pos; pos2++)
			type += *pos2;
		
		// ...And this is the function name
		while( (isalpha(*pos)||*pos=='_') && pos!='\0' ) {
			name += *pos;
			pos++;
		}
	}
//}}}
//{{{
	void emit_prototype(FILE *FP)
	{
		if(is_function)
			fprintf(FP,"%s%s%s", type.c_str(), name.c_str(), paramlist.c_str());
		else
			fprintf(FP, "%s", proto.c_str());
	}
//}}}
//{{{
	void emit_pointer(FILE *FP)
	{
		if(is_function)
			fprintf(FP,"%s(*%s)%s", type.c_str(), name.c_str(),
			                        paramlist.c_str());
		else
			fprintf(FP, "%s", proto.c_str());
	}
//}}}
//{{{
	void emit_instance(FILE *FP)
	{
		if(is_function)
			fprintf(FP, "\t%s,\n", name.c_str());
		else
			fprintf(FP, "\t%s", proto.c_str());
	}
//}}}
//{{{
	void emit_define(const char *structure, FILE *FP)
	{
		if(is_function)
			fprintf(FP,"#define %s (%s->%s)\n", name.c_str(), structure,
			                                    name.c_str());
		else
			fprintf(FP, "%s", proto.c_str());
	}
//}}}
	std::string getName(void) { return name; }
	
private:
	bool is_function;
	std::string proto;
	std::string type, name, paramlist;
};
//}}}
//{{{
int main(int argc, char **argv)
{
	std::vector<FunctionDefinition> functions;
	char prototype[512];
	unsigned ii;
	
	out = stdout;
	
	while(!feof(stdin)) {
		std::fgets(prototype, 512, stdin);
		functions.push_back( FunctionDefinition(prototype) );
	}
	
	// Boilerplate
	std::fputs("/* Generated file - do not edit */\n", out);
	std::fputs("#ifndef PROTO_H\n", out);
	std::fputs("#define PROTO_H\n\n", out);
	
	// Prototypes themselves (but only where the functions actually exist)
	std::fputs("#ifdef PROTOTYPE_MAIN\n", out);
	for(ii=0; ii<functions.size(); ii++)
		functions[ii].emit_prototype(out);
	std::fputs("#endif //PROTOTYPE_MAIN\n\n", out);
	
	// Definition of a big structure containing all the main-module functions
	// which should be exported
	std::fputs("#ifdef DLL_IMPORT_STRUCT\n", out);
	std::fputs("typedef struct import_functions {\n", out);
	for(ii=0; ii<functions.size(); ii++)
		functions[ii].emit_pointer(out);
	std::fputs("} import_functions;\n", out);
	std::fputs("#endif //DLL_IMPORT_PROTO\n\n", out);
	
	// Instance of a big structure containing all the main-module functions
	// which should be exported
	std::fputs("#ifdef DLL_IMPORT_INSTANCE\n", out);
	std::fputs("const import_functions body_functions = {\n", out);
	for(ii=0; ii<functions.size(); ii++)
		functions[ii].emit_instance(out);
	std::fputs("\tNULL\n", out);
	std::fputs("} import_functions;\n", out);
	std::fputs("#endif //DLL_IMPORT_PROTO\n\n", out);
	
	// #define statements to make function calls transparently use the big
	// structure
	std::fputs("#ifdef DLL_EXPORT_USAGE\n", out);
	for(ii=0; ii<functions.size(); ii++)
		functions[ii].emit_define("dll_body_functions", out);
	std::fputs("#endif //DLL_EXPORT_USAGE\n\n", out);
	
	// Ending boilerplate
	std::fputs("\n", out);
	std::fputs("#endif\n", out);
	std::fputs("\n", out);
	
	return 0;
}
//}}}

